open Glbdefs
open Genlex
module L=List
module Lin=Linexp
module C=Coeff
module M=Map
module P=Printf
module S=String

(* implement as a map from strings into linexp *)
module Lmap = M.Make (
  struct
    type t=int
    let compare s1 s2 = s1 -s2
  end
)

type t = 
    { 
      mutable vars :intset; (* the  variables that are involved. *)
      mutable eqmap: Lin.t Lmap.t ; (* The map for each dependenent variable to a denominator term and a linear expression *)
      mutable empty : bool (* Is this map empty? *)
    }



let universal vSet = { 
  vars = vSet;
  eqmap = Lmap.empty;
  empty = false 
}

let universal_eq () = universal Intset.empty
let empty_eq () = {
  vars = Intset.empty;
  eqmap = Lmap.empty;
  empty = true
}

let is_empty e = e.empty

let is_universal e = 
  match e.empty with 
    | true -> false
    | false -> (Lmap.is_empty e.eqmap)

let normalize_expr_with_term l s l1 = 
  (* repace any occurrence of s in l with 1/d * l1 *)
  let c= Lin.get_coefficient  s l in
  let l2= (* l2 = l - c* s + c * l1 *)
    if not (C.is_zero c)  
    then  (* remove the term corresponding to s in l3 *)
      let l3 = Lin.subtract_term s c l in 
	Lin.add_linexp l3 ( Lin.multiply_linexp c l1)
    else  l 
  in
    begin (* The coefficient of s in l2 should be zero *)
      assert (C.eq (Lin.get_coefficient s l2) C.zero);
      l2
    end


let normal_form em l = (* replace each occurrence of a dependent variable in l with an equivalent expression *)
  match em.empty with 
    | true -> l (* ignore call if set is empty *)
    | false -> begin
	let retexp = ref (Lin.const_expr (C.zero)) in
	let get_replacement_exp s = 
	  try
	    Some (Lmap.find s em.eqmap)
	  with
	    | Not_found -> None in
	let handle_term s c = match get_replacement_exp s with
	  | None -> (retexp := (Lin.add_term s c !retexp) )
	  | Some l -> retexp := Lin.add_linexp !retexp (Lin.multiply_linexp c l)
	in
	  Lin.iter handle_term l;
	  !retexp
      end

       


let add_equation em l =
  match em.empty with
    | true -> () (* ignore if map is empty *)
    | false ->  (* if map is not empty, then compute normal form of equation *)
	let l1 = normal_form em l in
	  if (Lin.is_constant l1) (* is the normal form a constant ?*)
	  then 
	    if (C.is_zero (Lin.get_constant l1) )  (* =0? *)
	    then () (* Ignore *)
	    else  (* adding 1 = 0; make empty *)
	      begin
		em.empty <- true;
		em.eqmap <- Lmap.empty;
	      end
	  else 
	    (*
	     * now choose the smallest variable 
	     * with non-zero coefficient in l as the
	     * new dependent variable
	     *)
	    let (dv,c) = (
	      try (* choose the smallest term in l1 *) 
		Lin.get_smallest_term l1 
	      with 
		| Not_found -> raise ( Failure ( "l1 cannot be a constant term. something is wrong. "))
	    ) 
	    in	    
	    let rem1 = Lin.divide_linexp c (Lin.subtract_term dv c l1) in (* pair dv,rem needs to be added *)
	    let rem = Lin.multiply_linexp C.minus_one rem1 in
	    let retmap = ref Lmap.empty in (* start an empty map *)
	    let handle_pair s le = (* normalize the expression l w.r.t rem in the map *) 
	      retmap := Lmap.add  s (normalize_expr_with_term le dv rem) !retmap in
	      begin
		em.vars <- Intset.union em.vars (Lin.get_variable_set l);
		Lmap.iter handle_pair em.eqmap;
		retmap := Lmap.add dv rem !retmap; (* add the new pair to the map *)
		em.eqmap <- !retmap;
		em.empty <- false
	      end
		
		
let set_to_zero ss = 
  let e = universal_eq() in
  let handle_str s = let l = Lin.expr_of C.zero s in 
    add_equation e l
  in
    Intset.iter handle_str ss ;
    e


let printVar i = "C("^(string_of_int i)^")"

let to_string em =
  match em.empty with
    | true -> "1=0;"
    | false -> 
	if not (is_universal em) then 
	  begin
	    let retstr = ref "" in
	    let hVar i = 
	      retstr := !retstr^(string_of_int i)^","
	    in
	      retstr := !retstr^"Vars: { ";
	      Intset.iter hVar em.vars;
	      retstr := !retstr^"}\n";
	    let print_term s l = 
	      retstr:= (!retstr^"\t C("^(string_of_int s)^") = "^(Lin.to_string printVar l)^" ; \n") 
	    in (
		Lmap.iter print_term em.eqmap;
		!retstr
	      )
	  end
	else "0=0;"



(* 
 * parser for equations. very similar to the 
 *  the cpoly parser 
 *)

let pCtx str =  (* String names are "C(..)" *)
  let n = String.length str in 
  let str' = String.sub str 1 (n - 2) in 
      try 
	(int_of_string str')
      with 
	| Failure s -> (
	    P.printf "Error: string %s is not a recognized variable name. Currently this parser uses x<variableNum> \n" str;
	    raise (Failure s)
	  )


let lex_tokens = [";";"="]@Lin.lex_tokens

let rec parse_eqs = parser 
  | [< l = parse_constr ; p1 = parse_more_eqs >] -> 
      begin
	add_equation p1 l;
	p1
      end
  | [< >] -> universal_eq () 
and parse_more_eqs = parser
      | [< l = parse_constr ; p1 = parse_more_eqs >] -> 
      begin
	add_equation p1 l;
	p1
      end
      | [< >] -> universal_eq () 
and parse_constr = parser
  | [< l1 = Lin.parse_linexp pCtx ; 'Kwd "="; l2 = Lin.parse_linexp pCtx; 'Kwd ";" >] -> 
      Lin.subtract_linexp l1 l2

let of_string s = 
  let lexer = make_lexer lex_tokens in 
  let parsec = parser 
    | [< e = parse_eqs; _ = Stream.empty >] -> e
  in
    parsec (lexer (Stream.of_string s))
      



	
let iter  func eq = 
  match eq.empty with 
    | true -> () (* do nothing if empty *)
    | false -> 
	let handle_a_term s l =  (* create a linear expression and send it to the function *)
	  let l1 = Lin.add_term s (C.mone) l in func l1
	in 
	  Lmap.iter handle_a_term eq.eqmap
	   
let map eq func = 
  match eq.empty with 
    | true -> []
    | false ->
	let retlst = ref [] in 
	let handle_one_term s l = 
	  let l1 = Lin.add_term s (C.minus_one) l in 
	    retlst := (func l1)::!retlst 
	in
	  begin
	    Lmap.iter handle_one_term eq.eqmap;
	    !retlst
	  end


let intersection e1 e2 = 
  match (e1.empty || e2.empty) with 
    | true -> empty_eq()
    | false -> 
	let eres = universal_eq () in
	let handle_exp l = add_equation eres l in
	  begin 
	    iter  (handle_exp) e1;
	    iter  (handle_exp) e2;
	    eres
	  end

exception Non
	    
let includes e1 e2 = (* for every equation in e1, check if normal form with e2 is ZERO *)
match (e1.empty,e2.empty) with 
  | (_,true) -> true 
  | (true,false) -> false 
  | (false,false) -> begin 
      let check_equation l  =  
	let l1 = normal_form e2 l in
	  if (Lin.is_zero l1) then () else raise Non
      in
	try
	  iter  check_equation e1;
	  true
	with 
	  | Non -> false

    end


(* Implement the join algorithm *)

let collect_independent_variables e = 
  match e.empty with
    | true -> Intset.empty
    | false -> 
	let depSet = ref Intset.empty in
	let handle_term s l = 
	  depSet := Intset.add s !depSet 
	in begin
	    Lmap.iter (handle_term) (e.eqmap);
	    Intset.diff e.vars !depSet
	  end

let clone e = {vars = e.vars; 
	       eqmap = e.eqmap;
	       empty= e.empty}

let get_variable_set e = e.vars



let dualize_var s = s


let dualize_vars_set vset =  (* dualize each variable in the set. Why is there no Set.S.map? *)
  let retset = ref Intset.empty in 
  let add_one_v s = retset:= Intset.add (dualize_var s) !retset in
    begin
      Intset.iter add_one_v vset;
      !retset
    end


let primalize_var s = s

let csig_ = Lin.const_sig
 
let dualize_eq e1 csig dvar  is_primal =  (* dualize an equation system e1 with constant signature csig and 
					     function dvar to translate variables into their duals *)

match e1.empty with
  | true -> universal_eq ()
  | false -> 
      if not (is_universal e1) then
	let idvars = collect_independent_variables e1  in (* collect the dependent and independent variables in e1 *)
	let vert = (* handle the affine part by creating a vertex *)
	  begin (* get_vertex_exp *)
	    let retexp = ref (Lin.const_expr C.zero) in
	    let handle_term s l = (* get the constant in l and add c * d_s *)
	      let c = Lin.get_constant l in 
		retexp := Lin.add_term (dvar s) c !retexp in
	      begin
		Lmap.iter handle_term e1.eqmap ;
		retexp := Lin.add_term csig (C.one) !retexp ;
		!retexp
	      end
	  end in (* get_vertex_exp *)
	let get_line_ivar eq iv = (* get dual expression by setting ivar to 1 *)
	  begin (* get_line_ivar *)
	    let retexp = ref (Lin.const_expr C.zero) in
	    let handle_one_term s l = 
	      let c = Lin.get_coefficient iv l  in  
		retexp:= Lin.add_term (dvar s) c !retexp in 
	      begin
		Lmap.iter handle_one_term e1.eqmap;
		retexp := Lin.add_term (dvar iv) C.one !retexp;
		add_equation eq !retexp
	      end 
	  end in (* get_line_ivar *)
	let reteq = universal_eq () in 
	  begin (* main routine *)
	    if is_primal 
	    then add_equation reteq (vert)
	    else () ;
	    Intset.iter (get_line_ivar reteq) idvars;
	    reteq
	  end (* main routine *)
      else 
	empty_eq ()

let dualize e evars = 
  match e.empty with 
    | true -> universal_eq()
    | false -> begin
	let e1 = dualize_eq e csig_ dualize_var true in
	let add_zero_line s = add_equation e1 (Lin.expr_of C.one (dualize_var s)) in
	let () = Intset.iter add_zero_line evars  in
	  e1
      end


let primalize ed vrs= 
  match ed.empty with 
    | true -> universal_eq()
    | false -> begin
	let e1 = dualize_eq ed csig_ primalize_var false in
	let add_zero_line s = add_equation e1 (Lin.expr_of C.one (primalize_var s)) in
	let all_dvrs = dualize_vars_set vrs in (* all possible dual vars *)
	let occ_dvrs = ed.vars in (* ones that actually occur in ed *)
	let dvrs = Intset.diff all_dvrs occ_dvrs in 
	let () = Intset.iter add_zero_line dvrs in (* what does not occur in occ_dvrs needs to be zeroed out *)
	  e1
      end


let join e1 e2 = 
  match (e1.empty,e2.empty) with
    | (true, true ) -> empty_eq ()
    | (true, false) -> clone e2
    | (false, true) -> clone e1
    | (false,false) -> begin
	let vr1  = Intset.diff e1.vars e2.vars in
	let vr2 = Intset.diff e2.vars e1.vars in
	let vrs = Intset.union e1.vars e2.vars in
	let ed1 = dualize e1 vr2 in
	let ed2 = dualize e2 vr1 in
	  (* 	let () = P.printf " (%s) \n (%s) \n " (to_string ed1) (to_string ed2) in  *)
	let ed = intersection ed1 ed2 in
	  (* 	let () = P.printf " (%s) \n " (to_string ed) in *)
	  primalize ed	vrs
      end


let project_into eq rv = 
  match (is_empty eq) with 
    | true -> empty_eq()
    | false -> 
	if (not (is_universal eq)) then 
	  let elim_vars = Intset.diff eq.vars rv in
	  let extra_vars= Intset.diff rv eq.vars in
	  let vrs = Intset.union elim_vars extra_vars in
	  let deq = dualize eq vrs in
	  let peq = primalize deq rv in 
	    peq
	else
	  universal_eq()



let get_map_size mp = 
  let j = ref 0 in
  let handle_term s l = incr j in
    Lmap.iter handle_term mp;
    !j

let size eq = get_map_size eq.eqmap


let getSolutionFromIndVarAssignment eq lI = 
  let handle_term s l lRet= 
    let (c,d) = Linexp.dot_product l lI in 
      Linexp.add_term s (Coeff.add c d) lRet 
  in 
    Lmap.fold handle_term eq.eqmap lI

let getSolutions eq = 
  if (eq.empty) 
  then [] 
  else (
    let sols = ref [ ] in 
    let iVars = collect_independent_variables eq in 
    let handleConstTerms () = 
      let l0 = Linexp.const_expr Coeff.one in 
	sols := (getSolutionFromIndVarAssignment eq l0)::!sols
    in 
    let handleVariable i = 
      let l0 = Linexp.expr_of (Coeff.one) i in
      let l1 = Linexp.add_constant Coeff.one l0 in 
	sols := (getSolutionFromIndVarAssignment eq l1)::!sols
    in 
      handleConstTerms ();
      Intset.iter handleVariable iVars;
      !sols
      
  )
