(** Linear equations package. Supports abstract domain 
  operations such as intersection, join, and inclusion
  tests. Also part of the [Cpoly] implementation.

  Warning: Some methods have side effects. Please 
  use the clone method to avoid tears.
  
*)
open Glbdefs

type t
  (** the type of equalities *)

val universal: intset -> t

val universal_eq:  unit -> t
  (** the universal set 0 =0 *)

val empty_eq: unit -> t
  (** the equality 1 =0 *)

val is_universal: t -> bool 
  (** is the equation the universal system ? *)

val is_empty: t -> bool 
  (** is it empty ? *)
val set_to_zero: Glbdefs.intset -> t
  (** add the equation s = 0 for each [s] in the input set *)

val dualize_var: int -> int
  (** obtain the dual of a variable. Used in the [Transform] module. *)

val dualize_vars_set: Glbdefs.intset -> Glbdefs.intset 
  (** dualize each variable in the set *)

val primalize_var: int -> int
  (** obtain the primal version of a variable that has been dualized *)

val get_variable_set: t -> Glbdefs.intset 
  (** get all the variables present. *)
  
val normal_form: t-> Linexp.t -> Linexp.t
  (** Simplify an expression as much as possible using the equalities *)

val add_equation: t -> Linexp.t -> unit
  (** add an equality *)


val to_string: t -> string
  (** convert to a string *)

val lex_tokens: string list 
  (** lexical tokens for parsing *)

val of_string: string -> t 
  (** convert from a string *)

val clone: t -> t
  (** clone this *)

val iter: (Linexp.t -> unit)-> t -> unit
  (** iterate a function over the equalities present *)

val map: t-> (Linexp.t -> 'a) -> 'a list
  (** iterate over the equalities stored
    and collect over a list. Universal set
    produces the empty list and the empty
    set calls with linear expression 1=0.
  *)

val dualize: t-> Glbdefs.intset ->  t
  (** dualize a set of equations into a set of generators, 
    represented by a equation system over dual variables *)

val primalize: t-> Glbdefs.intset -> t
  (** primalize a dual system. *)

  
val intersection: t-> t-> t
  (** intersect two sets of equalities *)

val includes: t-> t->bool
  (** perform an inclusion test of two sets of equalities. Returns true if the first arg. 
      logically includes the second. false otherwise. *)


val join: t-> t-> t
  (** compute the join of two sets of equalities *)




val project_into: t -> Glbdefs.intset -> t
  (** project into a set of variables *)

val size: t -> int
  (** the size in terms of the number of equations *)

val getSolutions: t -> Linexp.t list
(** get the solutions to the equations *)
