open Genlex
open Glbdefs
open Gmp
module L=List
module M=Map
module S=String


module C=Coeff
exception Bug of string

(* 
 * A linexp is implemented as a map 
 * from integer variable IDs to coefficients
 *)

type t = C.t Intmap.t


(* the signature for the constant term is -1 *)

let const_sig = (-1)




(* Form an expression of a coefficient and a variable *)

let expr_of c s =
  if (C.is_zero c)
  then Intmap.empty
  else Intmap.add s c Intmap.empty

let const_expr c = 
  if (C.is_zero c) 
  then Intmap.empty 
  else Intmap.add const_sig c Intmap.empty


(* 
 * apply ifun to each entry and 
 * collect the results in a list 
 *)


let map_list ifun l = 
  let lref =  ref [] in 
  let concat_entry  s c= begin
    lref := (ifun s c)::(!lref)
  end in 
    begin
      Intmap.iter (concat_entry) l ; 
      L.rev !lref
    end

(*
 * is the map zero?
 *)

let is_zero l = Intmap.is_empty l

(*
 * Unmarshall an expression into a string 
 *)

let to_string ctx l = 
  let some = ref false in (* has something been printed *)
  let sgn_string (c:C.t) = 
	  if (not (!some)) (* has something been printed? *) 
  	then 
     begin (* no *)
		some :=true;
    " " (* skip printing the sign this once *)
 end
else  if (C.geq c C.zero) then " + " else " "
in (* If c is +ve add a + to it *) 
  let to_string_term (s:int) (c:C.t) = begin (* convert a term into a string *)
if (C.eq c C.zero) then 
  begin
   	assert (false) (* this should never be happening *)
  end
else begin
  if (s== const_sig)
  then (sgn_string c)^(C.to_string c) 
  else (sgn_string c)^(C.to_string c)^" * "^ (Glbdefs.getName ctx s)
    end
  end in  (* Convert each term to a string and merge *)
if (is_zero l) 
then " 0 " 
else  S.concat "" (map_list to_string_term l)


(* 
 * simplify an expression 
 *)


let  collect_variables  l = 
  let var_of_term s c =  s
  in   
    map_list var_of_term l (* Collect all variables by iterating  *)
	
let get_variable_set l = 
  let sset = ref (Intset.empty) in 
  let add_term s c = 
    if const_sig == s
    then ()
    else begin (* just a simple check to make sure that zero terms are not present *)
      assert (not (C.is_zero c)); 
      sset := (Intset.add s !sset ) 
    end
  in 
    Intmap.iter (add_term) l; !sset
      

let iter f l = Intmap.iter f l
    

let get_coefficient s l = 
  try 
    Intmap.find s l 
  with 
    | Not_found -> C.zero

let get_constant  = get_coefficient const_sig

let add_term_t s c l = 
  let ts = C.add c (get_coefficient s l) in  (* add c to what is already there *)
    if (C.is_zero ts)  
    then Intmap.remove s l (* If result of addition is zero then remove entry *) 
    else Intmap.add s ts l (* If result of addition is non-zero then replace entry *)


let add_term s c l =  
   if (C.is_zero c) 
   then l
   else add_term_t s c l
  
let subtract_term  s c l = 
  add_term s (C.negate c) l


let add_constant c l = add_term const_sig c l

(* perform an abstract operation op *)

let binop_linexp op l1 l2 = 
  let retmap = ref l1  in (* create a reference l1 *) 
  let accumulate_term s c = begin (* accumulate a term *)
    assert (not (C.is_zero c));
    retmap := op s c !retmap 
  end in  		
    Intmap.iter (accumulate_term) l2; !retmap 

let add_linexp l1 l2 = binop_linexp add_term l1 l2

let subtract_linexp l1 l2 = binop_linexp subtract_term l1 l2 


let multiply_linexp c (l:t) = 
  if (C.is_zero c) 
  then Intmap.empty
  else begin
    let retmap = ref Intmap.empty in 
    let accumulate_term s c1 = begin 
      assert (not (C.is_zero c));
      retmap := Intmap.add s (C.multiply c1 c) !retmap 
    end
    in 
      Intmap.iter accumulate_term l;!retmap 
  end

 let divide_linexp c l = 
   if (C.is_zero c) 
   then (raise (Failure " Division by zero error in Linexp.divide_linexp "))
   else multiply_linexp (C.divide C.one c) l


(* Is the expression zero *)

(* 
 * Collect all the numerator denominator pairs 
 *)

let collect_fractions l = 
  let cfrac s c  = (Q.get_num c, Q.get_den c) 
  in (* Collect the fractions in each term *)
    map_list cfrac l 


(* 
 * compute a positive normalizing constant 
 *)

let normalizing_constant l = 
  let fList = collect_fractions l in 
  let hFrac  g (n,d) = 
    Z.lcm g d
  in 
  let z = L.fold_left hFrac (Z.of_int 1) fList in 
    if (Z.cmp z Z.zero) < 0 
    then Z.neg z
    else z
 
let normalize l = 
  let zz = normalizing_constant l in 
    assert((  Z.cmp zz Z.zero) >  0);
    multiply_linexp (Q.from_z zz) l

    
(* Compare two maps *)
let compare l1 l2 = Intmap.compare (C.compare) l1 l2 

(* compute dot product *)
let dot_product l1 l2 = 
  (* for each term in l1, find the coefficient in l2 and
     compute the dot product. *)
  let retc = ref C.zero  in 
  let constTerms = ref C.zero in 
  let handle_a_term s c = 
    if (s = const_sig)
    then
      let c2 = get_constant l2 in 
	constTerms := C.multiply c c2 
    else
      let c1 = get_coefficient s l2 in 
	retc := C.add !retc (C.multiply c c1) 
  in
    Intmap.iter handle_a_term l1;
    (!retc, !constTerms)

let apply l1 l2 = 
  (* treat l1 as a valuation to variables and
     l2 as an expression *)
  let retc = ref C.zero in 
  let handle_a_term s c = 
    if (s == const_sig) then 
      retc := C.add !retc c 
    else 
      let c1 = get_coefficient s l1 in 
	retc:= C.add !retc (C.multiply c c1)
  in 
    Intmap.iter handle_a_term l2;
    !retc 

(* parser for expressions *)
let lex_tokens = [ "+"; "-" ; "*"; "/";"("; ")" ]

exception ParseError of string

let rec parse_linexp ctx = parser 
  | [< e = (parse_sum ctx)  >] -> e
and parse_sum ctx = parser 
  | [< e = parse_term ctx  ; e1 = parse_more_sum ctx >] -> add_linexp e e1  
  | [< 'Kwd "-"; e = parse_term ctx; e1 = parse_more_sum ctx >] -> subtract_linexp e1 e
and parse_more_sum ctx = parser
  | [< 'Kwd "+" ; e = parse_term ctx ; e1 = parse_more_sum  ctx >] -> add_linexp e e1
  | [< 'Kwd "-" ; e = parse_term ctx ; e1 = parse_more_sum ctx >] -> subtract_linexp e1 e
  | [< e= parse_term ctx; e1=parse_more_sum ctx >] -> add_linexp e e1 
  | [< >] -> const_expr C.zero
and parse_term ctx = parser
  | [< c = C.parse_coeff; e=parse_term_rest ctx c >] -> e
  | [< 'Ident s >] -> ( expr_of C.one ( ctx s) )
  | [< 'Kwd "(" ; e = parse_linexp ctx ; 'Kwd ")" >] -> e
and parse_term_rest  ctx c = parser
  | [< 'Kwd "*" ; 'Ident s >] ->( expr_of c (ctx s) )
  | [< >] -> const_expr c
  

let of_string ctx s =   let lexer = make_lexer lex_tokens in 
let parsec = parser 
  | [< e=(parse_linexp ctx); _=Stream.empty >] -> e 
in 
  parsec (lexer (Stream.of_string s))
    

  
let strip_constant l = Intmap.remove const_sig l
  
let strip_eq l1 l2 = 
  let l1s= strip_constant l1 in 
  let l2s = strip_constant l2 in 
    (compare l1s l2s) == 0

let strip_compare l1 l2 = 
  compare (strip_constant l1) (strip_constant l2)



exception Not_constant

let is_constant l =  (* just verify that the smallest term is a constant *)
  try 
    let checkConst s c = 
      if  (not (s = const_sig) ) && (not (C.is_zero c)) 
      then raise Not_constant
      else ()
    in 
      Intmap.iter checkConst l ;
      true
  with 
    | Not_constant -> false


let get_smallest_term l = 
  let cmpTerms s' c'  (s,c) = 
    if (s' >= s) 
    then (s',c')
    else (s,c) 
  in 
    Intmap.fold cmpTerms l ((-1),Coeff.zero) 

exception Exitfun

let is_well_behaved l =
  let handle_term s c =
    if not (C.is_well_behaved c) then 
      raise Exitfun
  in
    try 
      Intmap.iter handle_term l;
      true
    with 
      | Exitfun -> false


let transform fn l  = 
  let retexp = ref (const_expr C.zero) in 
  let handle_term (s:int) (c:Coeff.t) = 
    if (s = const_sig) 
    then   retexp := add_term s c (!retexp)
    else
      let (s1,c1) = fn s c in 
	retexp := add_term s1 c1 (!retexp)
  in
    iter handle_term l ;
    !retexp


  

let selectVars offs l =
  let retExpr = ref (const_expr Coeff.zero) in 
  let hTerm s c = 
    if (s >= offs) then 
      retExpr := add_term (s-offs) c !retExpr
  in 
    Intmap.iter hTerm l;
    !retExpr
