(** A linear expression package.

Operations include

1. marshalling,

2. unmarshalling

3. arithmetic operations

Implement a linexp as a hashtable of string and coeff.
Use a special signature string for representing the
constant term. *)

type t
(** A linexp is a map from strings to coefficients *)

val const_sig: int
(** the signature for the constant string *)

val expr_of: Coeff.t -> int -> t
(** [expr_of c s] constructs the linexp with term [c*s] *)

val const_expr: Coeff.t -> t
(** construct a constant expression *)

val is_zero: t -> bool
(** is the linexp zero? *)

val to_string: Glbdefs.context -> t -> string
(** convert expression to string *)

val collect_variables: t -> int list
(** collect variables into a list. Include the const_sig if present. *)

val get_variable_set: t -> Glbdefs.intset
(** collect a set of variables. Do not include cont_sig *)

val iter: (int -> Coeff.t -> unit) -> t -> unit
(** A term iterator for linear expressions. Note that iterator will
also be called for the constant term.*)

val transform: (int -> Coeff.t -> (int * Coeff.t ) ) -> t -> t
(** transform using a function. The constant term is not fed in.  *)

val get_coefficient: int -> t -> Coeff.t
(** get coefficient corresponding to [int].
Returns [Coeff.zero] if input string not present. *)

val get_constant: t -> Coeff.t
(** Get the constant term *)

val add_term: int -> Coeff.t -> t -> t
(** add a term to the linexp *)

val subtract_term: int -> Coeff.t -> t -> t
(** subtract a term *)

val add_constant: Coeff.t -> t -> t
(** add a constant *)

val add_linexp: t -> t -> t
(** add two linexps *)

val subtract_linexp: t -> t -> t
(** subtract second argument linexp from first *)

val multiply_linexp: Coeff.t -> t -> t
(** multiply linexp by constant *)

val divide_linexp: Coeff.t -> t -> t


val compare: t -> t -> int
(** use map compare function to compare *)

val dot_product: t -> t -> (Coeff.t * Coeff.t)
(** compute the dot product of two linexps. Separate dot product of the linear part with the product of the constants.*)

val apply: t -> t -> Coeff.t
(** [apply l1 l2] treats [l1] as a valuation map for
variables and [l2] as an expression. It then computes
the value of [l2] assuming valuation [l1] *)

val lex_tokens: string list
(** tokens to enable lexical analysis *)

val parse_linexp: (string -> int) -> Genlex.token Stream.t -> t
(** parse a token stream and obtain a linexp *)

val of_string: (string -> int) -> string -> t
(** parse a string and obtain a linexp *)

val strip_constant: t -> t
(** [strip_constant l] returns the same expression as [l] except
for the constant term which is zeroed *)

val strip_eq: t -> t -> bool
(** [ strip_eq l1 l2] compares [l1] and [l2] on all but the constant
term *)

val strip_compare: t -> t -> int
(** [strip_compare l1 l2] strips the constant terms off [l1]
and [l2], applies compare *)

val is_constant: t -> bool
(** Is this linexp a constant *)

val is_well_behaved: t -> bool
(** check coefficients to see if they are well behaved, used to avoid numerical instabilities for
inversion join. *)

val selectVars: int -> t-> t
  (** Given an integer offset, zero out all variables whose values are less than the offset and 
      subtract offset from the index of all variables whose values are more *)

val  get_smallest_term: t -> (int * Coeff.t)
(** This gives the term with the highest index that has a non-zero coefficient, actually. Used by the equalities manager to rewrite this variable into a expression in terms of the smaller variables *)
