(*pp camlp4o pa_extend.cmo *)
(* -- We are implementing multivariate polynomials as functors -- *)

(* -- we need to parameterize multivariate polynomials by the coefficient  *)
(* ring that implements at least the following functions ---               *)
module L = List
module A = Array
module P = Printf

open Genlex
exception Break_From_Loop of int
module type CoefficientRing = sig
  type t
  val zero: t
  val one : t
  val mone: t
  val minus_one: t
  val add: t -> t -> t
  val subtract: t -> t -> t
  val negate: t -> t
  val multiply: t -> t -> t
  val divide: t -> t -> t
  val compare: t -> t -> int
  val geq: t -> t -> bool
  val eq: t -> t -> bool
  val leq: t -> t-> bool
  val ge: t -> t -> bool
  val le: t -> t -> bool
  val neq: t -> t -> bool
  val to_string: t -> string
  val of_string: string -> t
  val of_int: int -> t
  val is_zero: t -> bool
end

module S = functor (C: CoefficientRing) -> 
struct
  (* - Monomials are represented as integer arrays -- *)
  type mon = int array
      
  let findRightMostNonZero m = 
    let n = ref ( (A.length m) -1) in
      try 
	while (!n >= 0) do
	  if not (m.(!n) = 0)
	  then ( raise ( Break_From_Loop (!n)));
	  decr n
	done;
	0
      with 
	| Break_From_Loop i -> i + 1

  let getPowerOf j m = 
    if ( j >=  A.length m)
    then 0
    else m.(j)

  (* -- Code for comparing two monomials -- *)
  let compare_monomials m1 m2 = 
    (* Just do a straightforward lex compare *)
    let n1 = A.length m1 in 
    let n2 = A.length m2 in 
      try 
	assert (n1 = n2);
	for i = (n1 -1) downto 0
	do
	  if ( m1.(i) > m2.(i)) 
	  then raise ( Break_From_Loop 1)
	  else if (m1.(i) < m2.(i))
	  then raise (Break_From_Loop (-1))
	done;
	0
      with | Break_From_Loop a -> a

  let degree m = 
    Array.fold_left (+) 0 m 


 let compare_monomials_old m1 m2 = 
   let n1 =  findRightMostNonZero m1 in 
   let n2 = findRightMostNonZero m2 in 
     (* Order then so that n <= n' *)
   let (n, n') = if (n1 <= n2) then (n1, n2) else (n2, n1) in 
     try
	assert( n <= n');
	for i = 0 to (n - 1) 
	do
	  if (m1.(i) < m2.(i)) 
	  then raise (Break_From_Loop (- 1))
	  else 
	    if (m1.(i) > m2.(i)) 
	    then raise (Break_From_Loop 1)
	done;
	if (n1 = n2) 
	  (* -- To do: this is not enough, need to account for n1 != n2 but rest of  *)
	  (* larger monomial is all zeros --                                         *)
	then 0
	else 
	  if (n1 < n2) 
	  then - 1
	  else 1
      with | Break_From_Loop x -> x

  module MonOrdering = struct
    type t = mon
    let compare  = Pervasives.compare
  end

  module MMap = Map.Make (MonOrdering)

  type t = { 
    mutable _n: int; (*-- number of variables ---*)
    mutable _m: C.t MMap.t; (*-- a map from monomials to coefficients --*)
  } 


  let get_coefficient p mo = 
    assert((Array.length mo) = p._n);
    try
      MMap.find mo p._m
    with 
      | Not_found -> C.zero

  let zero_monomial nVars = A.make nVars 0 

  let is_zero_monomial m = 
    let n = A.length m in 
      try 
	for i = 0 to (n - 1) 
	do
	  if not ( m.(i) = 0) then (raise (Break_From_Loop 0))
	done;
	true
      with | Break_From_Loop x -> false
	
  let monomial_to_string m = 
    let n = A.length m in 
    let str = ref "" in
    let first = ref true in 
      for i = 0 to (n - 1)
      do
	if (m.(i) > 0) 
	then begin
	  if not (!first) 
	  then str := !str^"*" ;
	  str := !str^(Glbdefs.varToString i);
	  if (m.(i) > 1)
	  then str:= !str^(P.sprintf "^%d" m.(i));
	  first:= false
	end 
      done;
      !str

  let set_power mo i p = mo.(i) <- p
  let add_power mo i p = mo.(i) <- mo.(i) + p

  let monomial_product p1 p2 = 
    let n1 = A.length p1 in 
    let n2 = A.length p2 in 
    let (n_max, n_min, p_max, p_min) = 
      if (n1 >= n2) 
      then (n1, n2, p1, p2) 
      else (n2, n1, p2, p1)
    in 
    let pRet = A.copy p_max in 
      for i = 0 to (n_min - 1) 
      do
	pRet.(i) <- pRet.(i) + p_min.(i)
      done;
      pRet

  let monomial_product_assign p p'= 
    assert( (A.length p) <= (A.length p'));
    let n = A.length p in 
      for i = 0 to (n - 1) 
      do
	p.(i) <- p.(i) + p'.(i)
      done

	
  let padZerosToMonomial mo nZeros = 
    let mo1 = A.make nZeros 0 in 
      A.append mo mo1

  let zero nVars = { _n = nVars; 
		     _m = MMap.empty
		   } 
    
  let is_zero p = (p._n = 0) || (MMap.is_empty p._m)

  let const nVars c = 
    let p = zero nVars in 
      if not (C.is_zero c) then (
	let mon = zero_monomial nVars in 
	  p._m <- MMap.add mon c p._m
      );
      p


  let of_monomial m = 
    let n = A.length m in 
    let p = zero n in 
      p._m <- MMap.add m C.one p._m;
      p

  let of_var nVars i = 
    let m = zero_monomial nVars in 
      m.(i) <- 1;
      of_monomial m


  let createMonomials nVars degLimit =
    let retList = ref [] in 
    let addOneTerm lst = 
      assert(List.length lst = nVars);
      let m = Array.of_list lst in (*- make a monomial -*)
      if  not (is_zero_monomial m) 
      then retList := m::!retList
    in 
    let rec createMon degSoFar lst = 
      assert (degSoFar >= 0 );
      assert (degSoFar <= degLimit);
      assert ((List.length lst ) <= nVars);
      if (List.length lst = nVars) 
      then (*-- we are done, make a monomial and create a new template variable --*)
	addOneTerm lst 
      else
	(
	  if (degSoFar = 0) 
	  then createMon degSoFar (0::lst)
	  else (
	    for i = 0 to (degSoFar)
	    do
	      createMon (degSoFar - i) (i::lst)
	    done
	  )
	)
    in 
      createMon degLimit [] ;
      !retList


	
  let one nVars = 
    const nVars (C.one)

  let is_const p = 
    let is_const_helper m = try
      ignore ( MMap.iter 
		 (fun k c ->
		    assert ( not (C.is_zero c));
		    if not (is_zero_monomial k) 
		    then raise (Break_From_Loop 1) 
		 )
		 p._m
	     );
      true
    with | Break_From_Loop _  -> false
    in 
      (p._n = 0) || (MMap.is_empty p._m) || (is_const_helper p._m)

  let addTermAssign (c, mo) p = 
    assert ((A.length mo) <= p._n);
    let mo' = 
      if ((A.length mo) < p._n) 
      then padZerosToMonomial mo (p._n - (A.length mo))
      else mo 
    in 
    let c' = try MMap.find mo' p._m with Not_found -> C.zero in 
    let c'' = C.add c c' in
      if not (C.is_zero c'') 
      then 
	p._m <- MMap.add mo' c'' p._m
      else
	p._m <- MMap.remove mo' p._m


  let iterForEachTerm f p = 
    MMap.iter (fun k c -> ( 
		 assert (not( C.is_zero c)); 
		 f (c, k) 
	       ) 
	      ) p._m

  let mapForEachTerm f p = 
    let p' = zero p._n in 
      MMap.iter (fun k c -> 
		   let (c', k') = f (c, k) in 
		     addTermAssign (c', k') p' ) 
	p._m ;
      p'

  let to_string p = 
    if (is_zero p) then "0" 
    else 
      begin 
	let str = ref "" in 
	  MMap.iter ( fun k c -> 
			assert ( not (C.is_zero c));
			let cStr  = 
			  if (is_zero_monomial k) 
			  then 
			    if (C.eq c C.one)
			    then "1"
			    else 
			      (C.to_string c) 
			  else (
			    if (C.eq c C.one) 
			    then "+"
			    else 
			      begin
				if (C.eq c C.mone)
				then "-"
				else
				  begin 
				    let dStr = if (C.geq c C.zero)
				    then "+"^(C.to_string c)
				    else (C.to_string c) 
				    in 
				      dStr^"*" 
				  end 
			      end
			  )
			in
			  str:= !str^cStr^(monomial_to_string k) 
		    ) p._m;
	  !str
      end
	

  let addAssign p p' = 
    iterForEachTerm (fun t -> addTermAssign t p) p'
      
  let subAssign p p' = 
    iterForEachTerm (fun (c, k) -> addTermAssign ((C.negate c), k) p) p'

  let copy p = 
    { _n = p._n; 
      _m = p._m
    } 

  let poly_fop fAssign p p' = 
    let pRet = copy p in 
      fAssign pRet p';
      pRet

  let add p p' =
    poly_fop addAssign p p'

  let sub p p' = 
    poly_fop subAssign p p'


  let scaleAssign (c, m) p =
    if (C.is_zero c) 
    then p._m <- MMap.empty
    else (
      assert ((A.length m) <= p._n);
      let m' = 
	if ( (A.length m) < p._n) 
	then padZerosToMonomial m (p._n - (A.length m))
	else m 
      in
      let scaleTerm m1 c1 nMap = 
	assert ( not (C.is_zero c1));
	let m2 = monomial_product m' m1 in
	let c'= C.multiply c c1 in
	  if not (C.is_zero c') 
	  then MMap.add m2 c' nMap
	  else assert false 
      in
      let nMap = MMap.fold scaleTerm p._m MMap.empty in
	p._m <- nMap
    )

  let scalePoly (c,m) p = 
    let pRet = copy p in 
      scaleAssign (c,m) pRet;
      pRet

  (*-- Polynomial multiply --*)
  let mul p p' =  (*-- 
		    not a very good multiply routine.. 
		    but this will have to do for now 
		    --*)
    assert (p._n = p'._n);
    let nVars = p._n in 
    let pRet = zero nVars in 
      iterForEachTerm ( fun (c,m) -> (
			  assert (not (C.is_zero c));
			  addAssign pRet (scalePoly (c,m) p) 
			)
		      ) p';
      pRet

      

 (*-- Polynomial partial derivative --*)
  let partial i p = 
    assert ( i >= 0);
    let nVars = p._n in 
    let pRet = zero nVars in
    let handleTerm (c,m) = 
      assert ( not (C.is_zero c));
      let k = getPowerOf i m in 
	if (k = 0) 
	then (* this term is constant *) ()
	else (
	  let m' = A.copy m in 
	    m'.(i) <- m.(i) - 1 ;
	    assert( m'.(i) >= 0);
	    let trm' = ( C.multiply (C.of_int k) c ),m' in
	      addTermAssign trm' pRet
	)
    in 
      if ( i < nVars ) 
      then iterForEachTerm handleTerm p;
      pRet

      

  let grad p = 
    let nVars = p._n in 
    let retList = ref [] in 
      for i = (nVars -1) downto 0  do
	retList := (partial i p)::!retList
      done;
      !retList


  let lexer = Genlex.make_lexer ["*";"+"; "^"; "-"; "/"; "("; ")" ]
    
  let parseVariable str = 
    let n = String.length str in 
    let str' = String.sub str 1 (n - 1) in 
      try 
	(int_of_string str')
      with 
	| Failure s -> (
	    P.printf "Error: string %s is not a recognized variable name. Currently this parser uses x<variableNum> \n" str;
	    raise (Failure s)
	  )


  let rec parse_polynomial nVars = parser 
    | [< 'Kwd "-"; (c, m) = parse_more_term nVars; p = parse_more_polynomial nVars >] -> ((C.negate c), m):: p
    | [< (c, m) = parse_more_term nVars; p = parse_more_polynomial nVars >] -> 
	(c, m):: p
  and parse_more_polynomial nVars = parser 
    | [< (c, m) = parse_term nVars ; p = parse_more_polynomial nVars >] -> 
	(c, m):: p
    | [< >] -> []
  and parse_term nVars = parser
    | [< 'Kwd "+"; (c, m) = parse_more_term nVars >] -> c, m
    | [< 'Kwd "-"; (c, m) = parse_more_term nVars >] -> (C.negate c), m
  and parse_more_term nVars = parser
    | [< m = parse_monomial_nz nVars >] -> (C.one), m
    | [< c = parse_coefficient; m = parse_more_monomial nVars >] -> (c, m)
  and parse_monomial_nz nVars = parser 
    | [< 'Ident str; pp = parse_power; m = parse_more_monomial nVars >] -> 
	let varNum = parseVariable str in
	  assert (varNum < nVars);
	  add_power m varNum pp;
	  m
  and parse_more_monomial nVars = parser
    | [< 'Kwd "*" ; m = parse_monomial_nz nVars >] -> m
    | [< >] -> zero_monomial nVars 
  and parse_power = parser 
    | [< 'Kwd "^"; 'Int n >] -> assert (n > 0); n
    | [< >] -> 1
  and parse_coefficient = parser 
    | [< 'Int n; n2 = parse_more_coeff n >] -> n2
  and parse_more_coeff num = parser
    | [< 'Kwd "/"; 'Int den >] -> ( 
	assert (not (den = 0));
	let c1 = C.of_int num in 
	let c2 = C.of_int den in 
	  C.divide c1 c2
      )
  | [< >] -> C.of_int num

  let of_string nVars str = 
    let lst = parse_polynomial nVars (lexer(Stream.of_string str)) in
    let p = zero nVars in 
      L.iter (fun m -> addTermAssign m p) lst;
      p
	


end


module MP= struct  
  
  include S(Coeff)
    
  let lieDerivative (vField:t list) (p:t) =
    assert ((List.length vField) = (p._n));
    let retP = zero p._n in 
    let gr = grad p in 
    let hRes pJ  mpJ= 
      addAssign retP (mul pJ mpJ) 
    in
      List.iter2 hRes gr vField;
      retP

  let fPreMonomial (trs: t list) ((c: Coeff.t), (m: mon)) =
    let n = List.length trs in 
    let termP = ref (const  n c) in   (* Other wise , compute pre of the monomial *)
      for i = 0 to (n -1) 
      do
	let dI =  m.(i) in 
	let trsI = try List.nth trs i with Failure _ -> assert false in 
	  for j = 0 to (dI -1)
	  do
	    termP := mul trsI !termP 
	  done;   
      done;
      !termP

  let fPre (trs: t list) (p:t) = 
    assert( (List.length trs) = p._n); (* The update better have as many vars as the polynomial *)
    let retP = zero p._n in  (* What I plan to return *)
    let hTerm (c,m) =  (* Iterate through the terms of the polynomial *)
      if (is_zero_monomial m)  (* If it is the constant term *)
      then addTermAssign (c,m) retP (* simply add it to the return *)
      else (
	let termP = fPreMonomial trs (c,m) in 
	addAssign retP termP
      )
    in 
      iterForEachTerm hTerm p;
      retP
    
end



