(*---
  structures to store the system variable
  --*)

open Glbdefs
open Mpoly

type mode = {
  name: string;
  modeID: int;
  dyn: MP.t list
}

type trs = {
  preMode: int;
  postMode: int;
  trID: int;
  upd: MP.t list
}

type t = {
  mutable modeHash: mode Strhash.t;
  mutable modes: mode list;
  mutable transitions: trs list
}

let global_sys = {
  modeHash = Strhash.create 100 ; 
  modes = [] ;
  transitions = []
}

let getModeID m = m.modeID
let getModeName m = m.name
let getTrsID tr = tr.trID
let getModeVectorField m = m.dyn

let getPrePostModeIDs tr = tr.preMode, tr.postMode
let getUpdate tr = tr.upd

let intMap2List ?(defaultID = false) nVars fMap = 
  let vField = ref [] in 
  for i = (nVars-1) downto 0
  do
    let p = 
      try 
	Intmap.find i fMap
      with | 
	  Not_found -> 
	    if (not defaultID) 
	    then  MP.zero nVars
	    else  (* Make a polynomial consisting of variable i *)
		MP.of_var nVars i
	     
    in 
    vField := p::!vField
  done;
  vField
  
let addMode s fMap  = 
  let nVars = Glbdefs.numVars () in 
  let vField = intMap2List nVars fMap in 
  let m = { 
    name = s;
    modeID = List.length global_sys.modes;
    dyn = !vField
  } in 
  Strhash.add global_sys.modeHash s m;
  global_sys.modes <- global_sys.modes@[m]

    
let findModeIDByName n = 
  let m = Strhash.find global_sys.modeHash n in 
    m.modeID

let findModeByName n = 
  Strhash.find global_sys.modeHash n 


let findModeByID i = 
  assert( i >= 0 && i < (List.length global_sys.modes));
  try 
    List.nth global_sys.modes i
  with | _ -> assert false

let getModeNameFromID i = 
  let m = findModeByID i in 
    m.name

let addTransition preID postID uMap = 
  let tid = List.length global_sys.transitions in 
  let tau = { 
    preMode = preID;
    postMode = postID;
    trID = tid;
    upd = ! ( intMap2List ~defaultID:true (Glbdefs.numVars ())  uMap )
  } in 
  global_sys.transitions <- global_sys.transitions@[tau]

  
let ppVField oc vList = 
  let nVars = List.length vList in 
    for i = 0 to (nVars -1)
    do
      let sVar = Glbdefs.varToString i in 
      let pI =
	try 
	  List.nth vList i 
	with | Not_found -> assert false 
      in 
	Printf.printf "\t %s <- %s; \n" sVar (MP.to_string pI) 
    done

let ppMode oc m = 
  Printf.fprintf oc "mode %s { \n " m.name ;
  Printf.fprintf oc "# Mode ID: %d \n" m.modeID;
  Printf.fprintf oc "\t dynamics { \n";
  ppVField oc m.dyn;
  Printf.fprintf oc "\t } \n } \n"


let ppTrs oc t = 
  Printf.fprintf oc "[] %s -> %s { \n" (getModeNameFromID t.preMode) (getModeNameFromID t.postMode);
  ppVField oc t.upd;
  Printf.fprintf oc "   } \n"


let prettyPrint oc = 
  (*-- Pretty print the transition system -- *)
  List.iter (ppMode oc) global_sys.modes;
  List.iter (ppTrs oc) global_sys.transitions



let forEachMode foo sy = 
  List.iter foo sy.modes

let forEachTrs foo sy = 
  List.iter foo sy.transitions

let getOutgoingTrs sy m = 
  List.filter ( fun tr -> tr.preMode = m.modeID) (sy.transitions) 

let getIncomingTrs sy m = 
  List.filter ( fun tr -> tr.postMode = m.modeID) (sy.transitions) 

let numModes sy = 
  List.length sy.modes

let getPrePostModeNames tr = 
  let mPre = findModeByID tr.preMode in 
  let mPost = findModeByID tr.postMode in 
    (mPre.name, mPost.name)
