(*---
  Vector space iterations using template polyhedra
    1. Given a system with modes, ODEs in each mode and transition relations.
    2. We will have a map of each mode and an associated template polynomial.
       2.1 We will have global constraints on the template polynomial.
    3. Iteration: 
        1. Inside each mode, refine basis by 
          (a0) build template
          (a) compute lie derivative of template polynomial
          (b) compute constraints so that lie derivative lies in the vector space.
          (c) recompute the basis w.r.t constraints.
        2. For each transition, refine basis by 
          (a0) build template for post-mode
          (a) compute pre-condition of the post-mode template
          (b) compute template for pre-mode
          (c) match up post-mode template with pre-mode.
  --*)

open Glbdefs

module T=TemplatePoly
module MP = Mpoly.MP
module Sys=System


type t = {
  sys: Sys.t;
  modeData: (MP.t list ) Inthash.t; (*--- the number of variables and the template vars 
					associated with each location --*)
  mutable wList: Intset.t 
}

let getDataForModeID aData mid = 
  try
    Inthash.find aData.modeData mid
  with 
    | Not_found -> assert false

let getDataForMode aData m = 
  let mid = Sys.getModeID m in 
    getDataForModeID aData mid

let getSolutions tp eq = 
  let solList = Eqs.getSolutions eq in 
  let sList = List.map (T.TP.instantiateParams tp) solList in 
    List.filter (fun ee -> (not (MP.is_const ee ))) sList

let myCtx mName  i = Printf.sprintf "v%d" i (*mName *) 


let processSolution sExpr = 
  let lExpr = ref (Linexp.const_expr Coeff.zero) in 
  let hTerm i c = 
    if not ( i = Linexp.const_sig)
    then 
      lExpr := Linexp.add_term  (i-1) c !lExpr 
  in 
    Linexp.iter hTerm sExpr;
    !lExpr
 

let printAbstractionForMode aData m = 
  let mpList = getDataForMode aData m in 
    (*--- 
      For each basis polynomial, 
      express its Lie derivative as a linear combination 
      ---*)
  let n = Glbdefs.numVars() in 
  let (ntpVars, tPoly) = T.TP.makeTP n mpList in 
  let vField = Sys.getModeVectorField m in 
  let vSet = Glbdefs.getIntegerSet 0 ntpVars in 
  let printSol idx sExpr =
    let solExpr = processSolution sExpr in 
    Printf.printf "\t %s " (myCtx  (Sys.getModeName m) idx);
    Printf.printf " <- %s \n" (Linexp.to_string (myCtx (Sys.getModeName m) ) solExpr ) 
  in 
  let handleBasisElt idx p = 
    (*- 
      1. Compute Lie derivative of p
      2. Promote it to a template polyhedron
      3. Match up terms with tPoly
      4. For the resulting equations, get the solutions.
      5. Use the solutions to express the polynomials.
      -*)
    let pDer = MP.lieDerivative vField p in 
    let eq = Eqs.universal vSet in 
      T.TP.encodeMembership ntpVars eq pDer tPoly;
      let solList =   Eqs.getSolutions eq in 
	assert ( ( List.length solList) >= 1);
	List.iter (printSol idx ) solList 
  in 
    Printf.printf "mode %s { \n dynamics { \n" (Sys.getModeName m); 
    Glbdefs.iterIndexed handleBasisElt mpList;
    Printf.printf " } \n\n"




let printAbstractionForTrs aData tr = 
  let (pre,post) = Sys.getPrePostModeIDs tr in 
  let preList = getDataForModeID aData pre in 
  let postList = getDataForModeID aData post in 
    (*--- 
      For each basis polynomial, 
      express its Lie derivative as a linear combination 
      ---*)
  let n = Glbdefs.numVars() in 
  let upd = Sys.getUpdate tr in 
  let (ntpVars, tPoly) = T.TP.makeTP n preList in 
  let vSet = Glbdefs.getIntegerSet 0 ntpVars in 
  let printSol idx  sExpr =
    let solExpr = processSolution sExpr in 
      Printf.printf "\t %s\' " (myCtx (Sys.getModeName (Sys.findModeByID post))  idx);
      Printf.printf " = %s \n" (Linexp.to_string ( myCtx  (Sys.getModeName (Sys.findModeByID pre)) ) solExpr) 
  in 
  let handleBasisElt idx p = 
    (*- 
      1. Compute Lie derivative of p
      2. Promote it to a template polyhedron
      3. Match up terms with tPoly
      4. For the resulting equations, get the solutions.
      5. Use the solutions to express the polynomials.
      -*)
    let pDer = MP.fPre upd p in 
    let eq = Eqs.universal vSet in 
      T.TP.encodeMembership ntpVars eq pDer tPoly;
      let solList =   Eqs.getSolutions eq in 
	assert ( ( List.length solList) >= 1);
	List.iter (printSol idx ) solList 
  in 
    Glbdefs.iterIndexed handleBasisElt postList



let initData (glbSys:Sys.t) degLimit = 
  let mData: (MP.t list Inthash.t) = Inthash.create 1000 in
  let nVars = Glbdefs.numVars () in 
  let handleMode m = 
    let mid = Sys.getModeID m in 
    let mList = MP.createMonomials nVars degLimit in
    let mpList = List.map (MP.of_monomial) mList in 
      Inthash.replace mData mid mpList
  in 
    Sys.forEachMode handleMode glbSys ;
    { 
      sys = glbSys; 
      modeData = mData ;
      wList = Glbdefs.getIntegerSet 0 (Sys.numModes glbSys)
    }



let iterMode aData m = 
  (* -- 
     1. Take current template for the mode.
     2. Take its Lie derivative
     3. Match up terms to get equalities.
     4. Project out fresh variables.
     5. Normalize the new template. 
     If the template changed in terms of number of variables, return true else return false
     ---*)
  let n = Glbdefs.numVars() in 
  let mid = Sys.getModeID m in 
  let mpList = getDataForMode aData m in 
  let (ntVars, tp) = T.TP.makeTP n mpList in 
  let tpDeriv = T.TP.lieDerivative (Sys.getModeVectorField m)  tp in 
  let vSet = Glbdefs.getIntegerSet 0 (2 * ntVars) in 
  let eq = Eqs.universal vSet in 
    T.TP.matchTerms ntVars eq tpDeriv tp;
    let nList = getSolutions tp eq in 
      if ( ( (List.length nList) =  (List.length mpList) )
	   || ( ( List.length nList) <= 0) )
      then (  Inthash.replace aData.modeData mid nList; 
	      false )
      else ( Inthash.replace aData.modeData mid nList;
	     true )


  

let iterTrs aData tr = 
    (* --- 
       1. Take current template for the postMode
       2. Take its precondition
       3. Match up terms with template for premode to get equalities.
       4. Project out the fresh variables
       5. Normalize the new template.
       If the template changed in terms of number of variables, return true else return false.
       --*)
  let n = Glbdefs.numVars() in 
  let (pre,post) = Sys.getPrePostModeIDs tr in 
  let mpListPre = try Inthash.find aData.modeData pre with | Not_found -> assert false in 
  let mpListPost = try Inthash.find aData.modeData post with | Not_found -> assert false in 
  let (ntVarsPre, tpPre) = T.TP.makeTP n mpListPre in 
  let (ntVarsPost, tpPost) = T.TP.makeTP n mpListPost in 
  let tpPrecond = T.TP.fPre (Sys.getUpdate tr)  tpPost in 
  let vSet = Glbdefs.getIntegerSet 0 (ntVarsPre + ntVarsPost) in 
  let eq = Eqs.universal vSet in 
    T.TP.matchTerms ntVarsPost eq tpPrecond tpPre;
    let nList = getSolutions tpPost eq in 
      if ( ( (List.length nList) =  (List.length mpListPost) )
	   || ( ( List.length nList) <= 0) )
      then (   Inthash.replace aData.modeData post nList; false )
      else ( Inthash.replace aData.modeData post nList;
	     true )




let addModeToWorkList aData mid = 
  aData.wList <- Intset.add mid aData.wList

let removeModeFromWorkList aData mid = 
  aData.wList <- Intset.remove mid aData.wList

let modeFix aData m = 
  let iCount = ref 0 in 
  while (iterMode aData m) 
  do
    incr iCount
  done;
    !iCount > 0

let iterate aData m = 
  let mChange = modeFix aData m in
  let iTrsList = Sys.getIncomingTrs aData.sys m in 
  let hTrs change tr = 
    let cVal = iterTrs aData tr in 
      change || cVal 
  in 
  let hasChanged = List.fold_left hTrs  mChange iTrsList in 
    if (hasChanged) 
    then (
      let oTrsList = Sys.getOutgoingTrs aData.sys m in 
      let hTrs tr = 
	let (_, pMode) = Sys.getPrePostModeIDs tr in 
	  addModeToWorkList aData pMode
      in 
	List.iter hTrs oTrsList;
	true
    ) else (
      removeModeFromWorkList aData (Sys.getModeID m);
      false
    )
	

let modeInWorkList aData m  = 
  let mid = Sys.getModeID m in 
    Intset.mem mid aData.wList

let printResult aData = 
  let hMode m = 
    let mList = getDataForMode aData m in 
      Printf.printf "@Mode -> %s {  \n" (Sys.getModeName m);
      let hBasisElt i p  =
	Printf.printf " v%d ====> %s \n" i (MP.to_string p)
      in  
	Glbdefs.iterIndexed hBasisElt mList;
	Printf.printf " } \n"
  in 
  let hTrs tr = 
    let preStr, postStr = Sys.getPrePostModeNames tr in 
      Printf.printf " ----- Transition %d ------- \n" (System.getTrsID tr);
      Printf.printf " [] %s -> %s \n "  preStr postStr;
      printAbstractionForTrs aData tr;
      Printf.printf " <<--- Transition %d --->> \n\n" (System.getTrsID tr)
  in 
    Printf.printf "   --- Mode basis elements --- \n";
    Sys.forEachMode hMode aData.sys ;
    Printf.printf " <<--- Mode basis elements --->> \n\n";
    Sys.forEachMode (printAbstractionForMode aData) aData.sys ;
    Sys.forEachTrs hTrs aData.sys
	 

let doIt sys degLim = 
  let aData = initData sys degLim in 
  let change = ref true in 
    while (!change)
    do
      change := false ;
      Sys.forEachMode ( function m -> 
			  ( 
			    if (modeInWorkList aData m)
			    then (
			      let cVal = iterate aData m in 
				change := !change || cVal
			    )
			  )
		      ) aData.sys;
    done;
    printResult aData 







    
    
